{% extends "!page.html" %}
{% block menu %}
{{ super() }}
<iframe src="https://ghbtns.com/github-btn.html?user=mahmoud&repo=glom&type=star&count=true&size=medium" frameborder="0" scrolling="0" width="160px" height="30px" style="margin-left: 23px; margin-top: 10px;"></iframe>
{% endblock %}
